#include <cstring>
#include <iostream>
#include <fstream>
#include <stdio.h>
#include <stdlib.h>
#include <algorithm>
#include <ctime>

#include "sbutils.h"
#include "sbpgp.h"
#include "sbcertvalidator.h"
#include "sbxmlcore.h"
#include "sbxmlsig.h"
#include "sbxmlsec.h"
#include "sbxmlutils.h"
#include "sbx509.h"
#include "sbmisc.h"

using namespace SecureBlackbox;

std::string UnQuote(std::string s)
{
    if (s.length() < 1)
        return s;

	s.erase(remove( s.begin(), s.end(), '\"' ), s.end());

    return s;
}

void OutputCanonicalizationMethod(TElXMLCanonicalizationMethod value)
{
	std::cout << "Canonicalization method: ";
	switch (value)
	{
		case xcmCanon:
		{
			std::cout << "Canonical" << std::endl;
			break;
		}
		case xcmCanonComment:
		{
			std::cout << "Canonical with comments" << std::endl;
			break;
		}
		case xcmExclCanon:
		{
			std::cout << "Exclusive canonical" << std::endl;
			break;
		}
		case xcmExclCanonComment:
		{
			std::cout << "Exclusive canonical with comments" << std::endl;
			break;
		}
		case xcmMinCanon:
		{
			std::cout << "Minimal" << std::endl;
			break;
		}
		default:
		{
			std::cout << "Unknown" << std::endl;
			break;
		}
	}
}

void OutputMACMethod(TElXMLMACMethod value)
{
	std::cout << "MAC method: ";
	switch (value)
	{
		case xmmHMAC_MD5: 
		{
			std::cout << "MD5" << std::endl;
			break;
		}
		case xmmHMAC_SHA1: 
		{
			std::cout << "SHA1" << std::endl;
			break;
		}
		case xmmHMAC_SHA224: 
		{
			std::cout << "SHA224" << std::endl;
			break;
		}
		case xmmHMAC_SHA256: 
		{
			std::cout << "SHA256" << std::endl;
			break;
		}
		case xmmHMAC_SHA384: 
		{
			std::cout << "SHA384" << std::endl;
			break;
		}
		case xmmHMAC_SHA512: 
		{
			std::cout << "SHA512" << std::endl;
			break;
		}
		case xmmHMAC_RIPEMD160: 
		{
			std::cout << "RIPEMD160" << std::endl;
			break;
		}
		default:
		{
		    std::cout << "Unknown" << std::endl;
 		    break;
		}
	}
}

void OutputSignatureMethod(TElXMLSignatureMethod value)
{
	std::cout << "Signature method: ";
	switch (value)
	{
		case xsmRSA_SHA1:
		{
			std::cout << "RSA SHA1" << std::endl;
			break;
		}
		case xsmRSA_SHA256:
		{
			std::cout << "RSA SHA256" << std::endl;
			break;
		}
		case xsmRSA_SHA384:
		{
			std::cout << "RSA SHA384" << std::endl;
			break;
		}
		case xsmRSA_SHA512:
		{
			std::cout << "RSA SHA512" << std::endl;
			break;
		}
		case xsmRSA_RIPEMD160:
		{
			std::cout << "RSA RIPEMD160" << std::endl;
			break;
		}
		case xsmDSS:
		{
			std::cout << "DSS" << std::endl;
			break;
		}
		default:
		{
			std::cout << "Unknown" << std::endl;
			break;
		}
	}
}

void OutputSigMethodType(TElXMLSigMethodType value)
{
	std::cout << "Signature method type: ";
	if (value == xmtMAC)
		std::cout << "HMAC method" << std::endl;
	else
		std::cout << "Signature method" << std::endl;
}

void OutputKeyName(std::wstring value)
{
	std::wcout << L"Key name: " << value << std::endl;
}

void OutputSignatureTypes(TElXMLSignatureTypes value)
{
	std::cout << "Signature types: ";
	std::string s;
	if ((value & f_xstDetached) > 0)
		s += "Detached, ";
	if ((value & f_xstEnveloping) > 0)
		s += "Enveloping, ";
	if ((value & f_xstEnveloped) > 0)
		s += "Enveloped, ";

	std::cout << s << std::endl;
}

void LoadCertificate(TFileStream &F, std::string Password, TElXMLKeyInfoX509Data& X509KeyData)
{
	F.set_Position(0);
	TElX509Certificate cert(NULL);
	int k = cert.LoadFromStreamAuto(F, Password, 0);
	if (k == 0)
	{
		X509KeyData.set_Certificate(cert);
		cert.updateHandle(NULL); // detach wrapped class handle from wrapper class, so it won't be freed
	}
}


void DoVerify(TElXMLDOMElement &sigElement, std::string keyFile, std::string passphrase)
{
	std::time_t t = std::time(0);
    TElXMLVerifier Verifier (NULL);
    TElXAdESVerifier XAdESVerifier (NULL);
	TElXMLKeyInfoHMACData HMACKeyData(true);
	TElXMLKeyInfoRSAData RSAKeyData(true);
	TElXMLKeyInfoX509Data X509KeyData(true);
	TElXMLKeyInfoPGPData PGPKeyData(true);
	try
	{
		Verifier.set_XAdESProcessor(XAdESVerifier);
		Verifier.Load(sigElement);

		TElXMLReferenceList* Refs = Verifier.get_References();

		for (int i = 0; i < Refs->get_Count(); i++) // this code is unnecessary, it shows how to process application specific URI
		{
			TElXMLReference* Ref = Refs->get_Reference(i);
			if (Ref->get_URINode() == NULL)
			{
				std::wstring uri;
				Ref->get_URI(uri);
				if (uri == L"")
					Ref->set_URINode(sigElement.get_OwnerDocument()->get_DocumentElement());
				else
				{
					try
					{
						std::wstring id;
						ExtractIdFromLocalURI(uri, id);
						if (id != L"")
						{
							TElXMLDOMElementHandle hElement = FindElementById(sigElement.get_OwnerDocument()->get_DocumentElement(), id);
							TElXMLDOMElement Element(hElement, false);
							Ref->set_URINode(Element);
						}
					}
					catch (SBException E)
					{
						std::cout << "Error: " << E.what() << std::endl;
					}
				}
			}
		}

		TElXMLSigMethodType sigMethodType = Verifier.get_SignatureMethodType();
		OutputCanonicalizationMethod(Verifier.get_CanonicalizationMethod());
		OutputSigMethodType(sigMethodType);
		if (sigMethodType == xmtMAC)
			OutputMACMethod(Verifier.get_MACMethod());
		else
			OutputSignatureMethod(Verifier.get_SignatureMethod());

		OutputSignatureTypes(Verifier.get_SignatureType());
		std::wstring keyName;
		Verifier.get_KeyName(keyName);
		OutputKeyName(keyName);

		if (!Verifier.ValidateSignature())
		{
			// KeyInfo doesn't contain correct public key for the signature or data corrupted
			// trying to load key from keyFile
			if (keyFile == "")
			{
				std::cout << "Key not loaded." << std::endl;
				std::cout << "Signature is INVALID" << std::endl;
				return;
			}

			if (sigMethodType == xmtMAC)
			{
				TFileStream stream(keyFile, filemodeOpenRead);

				std::vector<uint8_t> Buf((size_t)stream.get_Size());
				if (stream.get_Size() > 0)
					stream.Read(&Buf.front(), (int32_t)stream.get_Size());

				TElHMACKeyMaterial* key = HMACKeyData.get_Key();
				key->set_Key_ByteArray(Buf);
				Verifier.set_HMACKey(HMACKeyData);
			}
			else
			{
				TElRSAKeyMaterial* key = RSAKeyData.get_RSAKeyMaterial();
				key->set_Passphrase(passphrase);

				TFileStream stream(keyFile, filemodeOpenRead);

				std::vector<uint8_t> Buf((size_t)stream.get_Size());
				if (stream.get_Size() > 0)
					stream.Read(&Buf.front(), (int32_t)stream.get_Size());

				try
				{
					key->LoadPublic(&Buf.front(), (int32_t)stream.get_Size());
				}
				catch (SBException E) {}

				if (!key->get_PublicKey())
				{
					stream.set_Position(0);
					try
					{
						key->LoadSecret(&Buf.front(), (int32_t)stream.get_Size());
					}
					catch (SBException E) {}
				}

				if (!key->get_PublicKey())
				{
					stream.set_Position(0);
					LoadCertificate(stream, passphrase, X509KeyData);
				}

				if (!key->get_PublicKey() && (X509KeyData.get_Certificate() == NULL))
				{
					TElPGPPublicKey pubKey;
					try
					{
						stream.set_Position(0);
						pubKey.LoadFromStream(stream);
						// if loaded ok
						PGPKeyData.set_PublicKey(pubKey);
						pubKey.updateHandle(NULL); // detach wrapped class handle from wrapper class, so it won't be freed
					}
					catch (SBException E) {}

					if (PGPKeyData.get_PublicKey() == NULL)
					{
						TElPGPSecretKey secKey;
						secKey.set_Passphrase(passphrase);
						try
						{
							stream.set_Position(0);
							secKey.LoadFromStream(stream);
							// if loaded ok
							PGPKeyData.set_SecretKey(secKey);
							secKey.updateHandle(NULL); // detach wrapped class handle from wrapper class, so it won't be freed
						}
						catch (SBException E) {}
					}
				}

				if (RSAKeyData.get_RSAKeyMaterial()->get_PublicKey())
					Verifier.set_KeyData(RSAKeyData);
				else if (X509KeyData.get_Certificate() != NULL)
					Verifier.set_KeyData(X509KeyData);
				else if ((PGPKeyData.get_PublicKey() != NULL) || (PGPKeyData.get_SecretKey() != NULL))
					Verifier.set_KeyData(PGPKeyData);
				else
				{
					std::cout << "Key not loaded." << std::endl;
					return;
				}
			}

			if (!Verifier.ValidateSignature())
			{
				std::cout << "Signature is INVALID" << std::endl;
				return;
			}
		}

		// signature is valid
		// validate a certificate used for signature validation
		if (Verifier.get_SignerCertificate() != NULL)
		{
			TElX509CertificateValidator CertificateValidator(NULL);
			TSBCertificateValidity Validity;
			TSBCertificateValidityReason Reason;
			TElXMLKeyInfoData* keyData = Verifier.get_SignerKeyData();
			if ((keyData != NULL) && (keyData->ClassNameIs_Inst("TElXMLKeyInfoX509Data")))
			{
				TElXMLKeyInfoX509Data X509(keyData->getHandle(), false);
				CertificateValidator.Validate(Verifier.get_SignerCertificate(), X509.get_CertStorage(), false, t, Validity, Reason);
			}
			else
				CertificateValidator.Validate(Verifier.get_SignerCertificate(), NULL, false, t, Validity, Reason);

			if (Validity == cvSelfSigned)
				std::cout << "The selected signature is signed by self-signed certificate which was not previously trusted" << std::endl;
			else if (Validity != cvOk)
			{
				//SigOK = false;
				if (Validity == cvChainUnvalidated)
					std::cout << "Signing certificate chain could not be validated completely." << std::endl;
				else
					std::cout << "Signing certificate is not valid." << std::endl;
			}
		}

		// validate references
		if (Verifier.ValidateReferences())
			std::cout << "Signature and references validated successfully." << std::endl;
		else
		{
			std::cout << "Signature validated successfully. But some references are not valid." << std::endl;
			TElXMLReferenceList* Refs = Verifier.get_References();
			for (int i = 0; i < Refs->get_Count(); i++)
			{
				std::cout << "Referenc #" << (i + 1) << std::endl;
				std::wstring uri;
				Refs->get_Reference(i)->get_URI(uri);
				std::wcout << L"  URI: " << uri << std::endl;
				if (Verifier.ValidateReference(Refs->get_Reference(i)))
					std::cout << "  Valid" << std::endl;
				else
					std::cout << "  INVALID" << std::endl;
			}
		}
	}
	catch (SBException E)
	{
		std::cout << "Error: " << E.what() << std::endl;
	}
}

int main(int argc, char **argv)
{
	int i;
	std::string helpMessage = "consoleverifier.exe [params]\n" \
		"[params]\n" \
		"-f : input file\n" \
		"-k : key filename\n" \
		"-p : passphrase\n";
	std::string xmlFilenameIn, keyFilename, keyPassword;

	try
	{
		#error Please pick the evaluation license key from <SecureBlackbox>\LicenseKey.txt file and place it here. If the evaluation key expires, you can request an extension using the form on https://www.eldos.com/sbb/keyreq/
SetLicenseKey("<put the key here>");
	}
	catch (SBException E)
	{
		std::cout << "Failed to set license key: " << E.what() << std::endl;
	}

	if (argc < 2)
	{
		std::cout << helpMessage << std::endl;
		return 0;
	}

	try
    {
		i = 1;
		while (i < argc)
		{
			std::string param;
			param.assign(argv[i]);

			if (param == "-f")
			{
                if (++i <= argc)
					xmlFilenameIn = UnQuote(argv[i]);				
			} else if (param == "-k")
			{
                if (++i <= argc)
					keyFilename = UnQuote(argv[i]);			
			} else if (param == "-p")
			{
                if (++i <= argc)
					keyPassword = UnQuote(argv[i]);			
			}

			i++;
		}
	}
	catch (...)
	{
		std::cout << "Failed to parse parameter" << std::endl;
		std::cout << helpMessage << std::endl;
		return 0;
	}

	TElXMLDOMDocument FXMLDocument;
	try
	{
		std::cout << "Loading XML document: " << xmlFilenameIn << std::endl;
		FXMLDocument.LoadFromFile(xmlFilenameIn);
	}
	catch (SBException E)
	{
		std::cout << "Failed to load XML document: " << E.what() << std::endl;
		return 0;
	}

	try
	{
		TElXMLNamespaceMap NSMap;
		NSMap.AddNamespace(L"ds", L"http://www.w3.org/2000/09/xmldsig#");
		TElXMLNodeSet NodeSet(FXMLDocument.SelectNodes(L"//ds:Signature", NSMap), true);
		if (NodeSet.get_Count() == 0)
		{
			std::cout << "No signatures found" << std::endl;
			return 0;
		}

		for (int i = 0; i < NodeSet.get_Count(); i++)
		{
			std::cout << "Checking signature #" << (i + 1) << std::endl;
			TElXMLDOMElement sigElement(NodeSet.get_Node(i)->getHandle(), false); // casting to TElXMLDOMElement
			DoVerify(sigElement, keyFilename, keyPassword);
			std::cout << std::endl;
		}
	}
	catch (SBException E)
	{
		std::cout << "Unexpected error: " << E.what() << std::endl;
		return 0;
	}

	return 0;
}